#!/usr/bin/perl

#    CatanAid 0.42
#
#    Copyright S. Schimkat <www.schimkat.dk>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA





# Loading required modules
BEGIN {

   my $mod_TK = eval { require Tk; };
   my $mod_DB = eval { require Data::Dumper; };
   my $mod_64 = eval { require MIME::Base64; };
   my $mod_CZ = eval { require Compress::Zlib; };
   my $mod_ST = eval { require strict; };
   my $mod_GD = eval { require GD; };

   if (!($mod_TK) or !($mod_DB) or !($mod_64) or !($mod_CZ) or !($mod_ST) or !($mod_GD)) {
      print "\nRequired modules not found. You need to install\nthe missing modules in order to use CatanAid:\n" . 
            "\nTK\t\t"           . (($mod_TK) ? 'OK' : 'Not found') . 
            "\nData::Dumper\t"   . (($mod_DB) ? 'OK' : 'Not found') .
            "\nMIME::Base64\t"   . (($mod_64) ? 'OK' : 'Not found') .
            "\nCompress::Zlib\t" . (($mod_CZ) ? 'OK' : 'Not found') .
            "\nstrict\t\t"       . (($mod_ST) ? 'OK' : 'Not found') .
            "\nGD\t\t"           . (($mod_GD) ? 'OK' : 'Not found') . "\n\nPress <enter> to exit ...";
      local( $| ) = 1;
      my $resp= <STDIN>;
      exit(0);
   }

}

use Tk;
use Data::Dumper;
use MIME::Base64;
use Compress::Zlib;
use strict;
use GD;

use subs qw(ShuffleArray GenerateLandscape DrawLandscape GenerateText);






# Defing fontdata array
my %fontdata = ( 'chars' => {'65'  => { pos => [1, 1]  },
                             '66'  => { pos => [1, 2]  },
                             '67'  => { pos => [1, 3]  },
                             '68'  => { pos => [1, 4]  },
                             '69'  => { pos => [1, 5]  },
                             '70'  => { pos => [1, 6]  },
                             '71'  => { pos => [1, 7]  },
                             '72'  => { pos => [1, 8]  },
                             '73'  => { pos => [1, 9]  },
                             '74'  => { pos => [1, 10] },
                             '75'  => { pos => [1, 11] },
                             '76'  => { pos => [1, 12] },
                             '77'  => { pos => [1, 13] },
                             '78'  => { pos => [2, 1]  },
                             '79'  => { pos => [2, 2]  },
                             '80'  => { pos => [2, 3]  },
                             '81'  => { pos => [2, 4]  },
                             '82'  => { pos => [2, 5]  },
                             '83'  => { pos => [2, 6]  },
                             '84'  => { pos => [2, 7]  },
                             '85'  => { pos => [2, 8]  },
                             '86'  => { pos => [2, 9]  },
                             '87'  => { pos => [2, 10] },
                             '88'  => { pos => [2, 11] },
                             '89'  => { pos => [2, 12] },
                             '90'  => { pos => [2, 13] },
                             '97'  => { pos => [3, 1]  },
                             '98'  => { pos => [3, 2]  },
                             '99'  => { pos => [3, 3]  },
                             '100' => { pos => [3, 4]  },
                             '101' => { pos => [3, 5]  },
                             '102' => { pos => [3, 6]  },
                             '103' => { pos => [3, 7]  },
                             '104' => { pos => [3, 8]  },
                             '105' => { pos => [3, 9]  },
                             '106' => { pos => [3, 10] },
                             '107' => { pos => [3, 11] },
                             '108' => { pos => [3, 12] },
                             '109' => { pos => [3, 13] },
                             '110' => { pos => [4, 1]  },
                             '111' => { pos => [4, 2]  },
                             '112' => { pos => [4, 3]  },
                             '113' => { pos => [4, 4]  },
                             '114' => { pos => [4, 5]  },
                             '115' => { pos => [4, 6]  },
                             '116' => { pos => [4, 7]  },
                             '117' => { pos => [4, 8]  },
                             '118' => { pos => [4, 9]  },
                             '119' => { pos => [4, 10] },
                             '120' => { pos => [4, 11] },
                             '121' => { pos => [4, 12] },
                             '122' => { pos => [4, 13] },
                             '49'  => { pos => [5, 1]  },
                             '50'  => { pos => [5, 2]  },
                             '51'  => { pos => [5, 3]  },
                             '52'  => { pos => [5, 4]  },
                             '53'  => { pos => [5, 5]  },
                             '54'  => { pos => [5, 6]  },
                             '55'  => { pos => [5, 7]  },
                             '56'  => { pos => [5, 8]  },
                             '57'  => { pos => [5, 9]  },
                             '48'  => { pos => [5, 10] },
                             '60'  => { pos => [5, 11] },
                             '45'  => { pos => [5, 12] },
                             '62'  => { pos => [5, 13] },
                             '44'  => { pos => [6, 1]  },
                             '46'  => { pos => [6, 2]  },
                             '59'  => { pos => [6, 3]  },
                             '58'  => { pos => [6, 4]  },
                             '95'  => { pos => [6, 5]  },
                             '92'  => { pos => [6, 6]  },
                             '33'  => { pos => [6, 7]  },
                             '34'  => { pos => [6, 8]  },
                             '35'  => { pos => [6, 9]  },
                             '164' => { pos => [6, 10] },
                             '37'  => { pos => [6, 11] },
                             '38'  => { pos => [6, 12] },
                             '47'  => { pos => [6, 13] },
                             '40'  => { pos => [7, 1]  },
                             '41'  => { pos => [7, 2]  },
                             '61'  => { pos => [7, 3]  },
                             '63'  => { pos => [7, 4]  },
                             '43'  => { pos => [7, 5]  },
                             '123' => { pos => [7, 6]  },
                             '125' => { pos => [7, 7]  },
                             '91'  => { pos => [7, 8]  },
                             '93'  => { pos => [7, 9]  },
                             '180' => { pos => [7, 10] },
                             '96'  => { pos => [7, 11] },
                             '39'  => { pos => [7, 12] },
                             '42'  => { pos => [7, 13] }});





# Setting up the Window
my %window = (width => 810, height => 630);

my $MW = MainWindow->new(-title => "CatanAid",
                         -background => "black");
my $canvas = $MW->Canvas(-background => "black",
                         -height             => $window{height},
                         -width              => $window{width},
                         -highlightthickness => 0)->pack(-side => 'top');





# Reading graphicsdata
open (GRAPHICS, '<graphicsdata.dat') or die "Can not find datafile.\n";
binmode (GRAPHICS);

my $zlib = inflateInit();
my $rawdata;
my $encodeddata;
my ($output, $status);
while (read(GRAPHICS, $rawdata, 4096)) {
   ($output, $status) = $zlib->inflate(\$rawdata);
   $encodeddata = $encodeddata . $output;
}

close GRAPHICS;
undef $rawdata;
undef $status;
undef $output;
undef $zlib;

my $name;
my %graphicsdata;
foreach (split (/\n/, $encodeddata)) {
   next if (/^\#/ or /^$/);
   if (/^\<(\w+)\/(\w+)\>$/) {
      $name = $1 . "_" . $2;
      next;
   }
   $graphicsdata{$name} .= $_ . "\n";
}

undef $encodeddata;

my $image;
my %graphics;
foreach $image (keys %graphicsdata) {
   $graphics{canvas}{$image} = $canvas->Photo(-data => $graphicsdata{$image});
   $graphics{GD}{$image}     = GD::Image->newFromGifData(decode_base64($graphicsdata{$image}));
   delete $graphicsdata{$image};
}

undef %graphicsdata;





# Reading landscapedata
my $landscape;
my %landscapedata;

foreach $landscape ('Standard', 'Expansion', 'Funny') {

   open (LANDSCAPE, '<' . $landscape . '.caa') or die "Can not open landscape datafile.\n";
   binmode (LANDSCAPE);

   $zlib = inflateInit();
   while (read(LANDSCAPE, $rawdata, 4096)) {
      ($output, $status) = $zlib->inflate(\$rawdata);
      $encodeddata = $encodeddata . $output;
   }
   close LANDSCAPE;

   $encodeddata =~ s/saved/$landscape/;
   eval($encodeddata);

}

undef $encodeddata;
undef $landscape;
undef $zlib;



# Drawing main interface
my $background = $canvas->createImage(0, 0, -image => $graphics{canvas}{interface_background},
                                            -anchor => 'nw');

my $standardbutton = $canvas->createImage(100, 525, -image  => $graphics{canvas}{interface_standardbutton},
                                                    -anchor => 'nw',
                                                    -tags   => [ 'standardbutton']);

my $expansionbutton = $canvas->createImage(20, 519, -image  => $graphics{canvas}{interface_expansionbutton},
                                                    -anchor => 'nw',
                                                    -tags   => [ 'expansionbutton']);

my $funnybutton = $canvas->createImage(52, 456, -image  => $graphics{canvas}{interface_funnybutton},
                                                -anchor => 'nw',
                                                -tags   => [ 'funnybutton']);

my $savebutton = $canvas->createImage(190, 540, -image  => $graphics{canvas}{interface_savebutton},
                                                -anchor => 'nw',
                                                -tags   => [ 'savebutton']);

my $loadbutton = $canvas->createImage(240, 540, -image  => $graphics{canvas}{interface_loadbutton},
                                                -anchor => 'nw',
                                                -tags   => [ 'loadbutton']);

my $generateimagebutton = $canvas->createImage(290, 554, -image  => $graphics{canvas}{interface_generateimagebutton},
                                                         -anchor => 'nw',
                                                         -tags   => [ 'generateimagebutton']);
my $landscapemode;

$canvas->bind(qw/standardbutton <Button-1>/ => sub {
   $landscapemode = "Standard";
   GenerateLandscape;
   DrawLandscape("canvas");
});

$canvas->bind(qw/expansionbutton <Button-1>/ => sub {
   $landscapemode = "Expansion";
   GenerateLandscape;
   DrawLandscape("canvas");
});

$canvas->bind(qw/funnybutton <Button-1>/ => sub {
   $landscapemode = "Funny";
   GenerateLandscape;
   DrawLandscape("canvas");
});

$canvas->bind(qw/savebutton <Button-1>/ => sub {

   my $filename = $MW->getSaveFile(-filetypes        => [['CatanAid Datafile', '.caa'], ['All files', '*']],
                                   -defaultextension => 'caa',
                                   -title            => 'Save landscape as ...');

   if ($filename ne '') {

      $Data::Dumper::Terse = 1;

      my $savedata = "\# CatanAid landscape savefile version 2\n\n\$landscapedata{saved} = " . Dumper($landscapedata{$landscapemode}) . ";\n";
      my $zlib = deflateInit(-Level => Z_BEST_COMPRESSION);

      open (SAVEFILE, ">$filename");
      binmode (SAVEFILE);

      my ($output, $status) ;
      ($output, $status) = $zlib->deflate($savedata) ;
      print SAVEFILE $output;
      ($output, $status) = $zlib->flush() ;
      print SAVEFILE $output ;

      close SAVEFILE;

      if (-e ($filename)) {

         $MW -> messageBox(-type    => "ok",
                           -icon    => "info",
                           -title   => "Landscape saved",
                           -message => "Landscape succesfully saved.");

      }
      else {

         $MW -> messageBox(-type    => "ok",
                           -icon    => "error",
                           -title   => "Error",
                           -message => "Could not save landscape as " . $filename);

      }

   }
});

$canvas->bind(qw/loadbutton <Button-1>/ => sub {

   my $filename = $MW->getOpenFile(-filetypes        => [['CatanAid Datafile', '.caa'], ['All files', '*']],
                                   -defaultextension => 'caa',
                                   -title            => 'Load landscape ...');

   if (-e $filename) {

      $landscapemode = "saved";

      if (open (SAVEFILE, "<$filename")) {
         binmode (SAVEFILE);

         my $zlib = inflateInit();
         my ($output, $status);

         my $data;
         my $savedata;
         while (read(SAVEFILE, $savedata, 4096)) {

           ($output, $status) = $zlib->inflate(\$savedata);
           $data = $data . $output;

         }
         close SAVEFILE;

         eval($data);
         DrawLandscape("canvas");

      }
      else {


         $MW -> messageBox(-type    => "ok",
                           -icon    => "error",
                           -title   => "Error",
                           -message => "Could not load landscape " . $filename);
  

      }      


   }
});

my $white;
my $black;
$canvas->bind(qw/generateimagebutton <Button-1>/ => sub {

   my $filename = $MW->getSaveFile(-filetypes        => [['CatanAid landscape', '.png'], ['All files', '*']],
                                   -defaultextension => 'png',
                                   -title            => 'Export landscape as ...');

   if ($filename ne '') {

      $image = new GD::Image($window{width}, $window{height}, 1);
      $image->copy($graphics{GD}{interface_background}, 0, 0, 0, 0, $window{width}, $window{height});

      $white = $image->colorAllocate(255,255,255);
      $black = $image->colorAllocate(0,0,0);

      DrawLandscape("image");

      my ($sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst) = localtime($landscapedata{$landscapemode}{information}{date});
      my $message = 'Landscape created by ' . $landscapedata{$landscapemode}{information}{user} . ', ' . $mday . '/' . ($mon + 1) . '-' . ($year + 1900);
      $image->copy( GetTextImage($message, 'goldbold', 16) , 30 , 575, 0 , 0 , 800 , 32);

      open (IMAGE, ">$filename");
      binmode IMAGE;
      print IMAGE $image->png();
      close IMAGE;

      if (-e ($filename)) {

         $MW -> messageBox(-type    => "ok",
                           -icon    => "info",
                           -title   => "Landscape saved",
                           -message => "Landscape succesfully exported.");

      }
      else {

         $MW -> messageBox(-type    => "ok",
                           -icon    => "error",
                           -title   => "Error",
                           -message => "Could not export landscape to " . $filename);


      }


   }


});





$landscapemode = "Standard";
GenerateLandscape;
DrawLandscape("canvas");

MainLoop();





sub ShuffleArray {

   my $array = shift;
   my $i;

   for ($i = @$array; --$i; ) {
      my $j = int rand ($i+1);
      next if $i == $j;
      @$array[$i,$j] = @$array[$j,$i];
   }

}





sub GenerateLandscape {

   # Shuffle arrays
   ShuffleArray (\@{$landscapedata{$landscapemode}{cordinates}{land}});
   ShuffleArray (\@{$landscapedata{$landscapemode}{cordinates}{port}});
   ShuffleArray (\@{$landscapedata{$landscapemode}{dicepoints}});

   # Assign landscapes
   my $j = 0;
   my $i = 0;
   my $landscape;
   foreach $landscape (keys %{$landscapedata{$landscapemode}{landscapes}}) {
      for ($j = 1; $j <= $landscapedata{$landscapemode}{landscapes}{$landscape}; $j++) {
         $landscapedata{$landscapemode}{cordinates}{land}[$i]{object} = "landscapes_" . $landscape;
         $i++;
      }
   }

   # Assign ports
   $i = 0;
   my $port;
   foreach $port (keys %{$landscapedata{$landscapemode}{ports}}) {
      for ($j = 1; $j <= $landscapedata{$landscapemode}{ports}{$port}; $j++) {
         $landscapedata{$landscapemode}{cordinates}{port}[$i]{object} = "ports_" . $port;
         $i++;
      }
   }

   # Assign landscape information
   my $username = 'Unknown user';
   $username = ($ENV{'USER'}) ? $ENV{'USER'} : $username;
   $username = ($ENV{'USERNAME'}) ? $ENV{'USERNAME'} : $username;
   $landscapedata{$landscapemode}{information}{date} = time();
   $landscapedata{$landscapemode}{information}{user} = ucfirst($username);


}





sub DrawLandscape {

   my $target = shift;

   # Clear current landscape
   our @CurrentObjects;
   while (($target eq "canvas") and ($#CurrentObjects > -1)) {
      my $object = shift @CurrentObjects;
      $canvas->delete($object)
   }

   # Create landscape
   my $i;
   my $landscapetype;
   foreach $landscapetype (sort {$b cmp $a} (keys %{$landscapedata{$landscapemode}{cordinates}})) {
      for ($i = 0; $i <= $#{$landscapedata{$landscapemode}{cordinates}{$landscapetype}}; $i++) {
         # print "Debug: Mode = $landscapemode : Landscapetype = $landscapetype : X = $landscapedata{$landscapemode}{cordinates}{$landscapetype}[$i]{x} : Y = $landscapedata{$landscapemode}{cordinates}{$landscapetype}[$i]{y} : Object = $landscapedata{$landscapemode}{cordinates}{$landscapetype}[$i]{object}\n";

         if ($target eq "canvas") {

            my $landscape = $canvas->createImage($landscapedata{$landscapemode}{cordinates}{$landscapetype}[$i]{x},
                                                 $landscapedata{$landscapemode}{cordinates}{$landscapetype}[$i]{y},
                                                 -image => $graphics{canvas}{$landscapedata{$landscapemode}{cordinates}{$landscapetype}[$i]{object}});
            push @CurrentObjects, $landscape;

         }
         elsif ($target eq "image") {

            my ($width, $height) = $graphics{GD}{$landscapedata{$landscapemode}{cordinates}{$landscapetype}[$i]{object}}->getBounds();
            $image->copy($graphics{GD}{$landscapedata{$landscapemode}{cordinates}{$landscapetype}[$i]{object}},
                         $landscapedata{$landscapemode}{cordinates}{$landscapetype}[$i]{x} - ($width / 2),
                         $landscapedata{$landscapemode}{cordinates}{$landscapetype}[$i]{y} - ($height / 2),
                         0,
                         0,
                         $width,
                         $height)

         }
      }
   }

   # Create dicepoints
   
   for $i (0 .. $#{$landscapedata{$landscapemode}{cordinates}{land}}) {

      if ($target eq "canvas") {

         my $number = $canvas->createText($landscapedata{$landscapemode}{cordinates}{land}[$i]{x},
                                          $landscapedata{$landscapemode}{cordinates}{land}[$i]{y}, -fill    => 'black',
                                                                                                   -text    => $landscapedata{$landscapemode}{dicepoints}[$i],
                                                                                                   -justify => 'center',
                                                                                                   -font    => 'Helvetica 8');
         push @CurrentObjects, $number

      }
      elsif ($target eq "image") {

         $image->string(gdSmallFont,
                        $landscapedata{$landscapemode}{cordinates}{land}[$i]{x} - ((length($landscapedata{$landscapemode}{dicepoints}[$i]) > 1) ? 6 : 3),
                        $landscapedata{$landscapemode}{cordinates}{land}[$i]{y} - 7,
                        $landscapedata{$landscapemode}{dicepoints}[$i],
                        $black);


      }

   }

}





sub GetTextImage {

   my $string   = shift;
   my $font     = shift;
   my $fontsize = shift;

   $font = 'fonts_' . $font;
   ($fontdata{fonts}{$font}{width}, $fontdata{fonts}{$font}{height}) = $graphics{GD}{$font}->getBounds();
   $fontdata{fonts}{$font}{width}  = $fontdata{fonts}{$font}{width}  / 15;
   $fontdata{fonts}{$font}{height} = $fontdata{fonts}{$font}{height} / 15;

   my $image = new GD::Image(800, $fontsize * 2, 1);

   my $black  = $image->colorAllocateAlpha(0, 0, 0, 127); 
   $image->transparent($black); 
   $image->alphaBlending(0);
   $image->filledRectangle(0,0,800,$fontsize * 2, $black); 

   my $char;
   my $charpos = 0;
   my @chararray = unpack("C*", $string);
   foreach $char (@chararray) {

      if (!($fontdata{fonts}{$font}{$fontsize}{$char}{char})) {

         my $srcX;
         my $srcW;

         my $charX;
         my $charY;

         if ($char == '32') {

            $srcX = 1;
            $srcW = $fontdata{fonts}{$font}{width} / 3;

         }
         else {

            $charX  = $fontdata{chars}{$char}{pos}[1] * $fontdata{fonts}{$font}{width};
            $charY  = (($fontdata{chars}{$char}{pos}[0] * $fontdata{fonts}{$font}{height}) * 2) - $fontdata{fonts}{$font}{height};
            my $scanX;
            my $scanY;
            for ($scanX = $charX; $scanX <= ($charX + $fontdata{fonts}{$font}{width}); $scanX++) { 
               for ($scanY = $charY; $scanY <= ($charY + $fontdata{fonts}{$font}{height}); $scanY++) {

                  my ($red, $green, $blue) = $graphics{GD}{$font}->rgb($graphics{GD}{$font}->getPixel($scanX, $scanY));

                  if (($srcX == 0) and (($red + $green + $blue) > 0)) {
                     $srcX = $scanX;
                  }

                  if (($srcX > 0) and (($red + $green + $blue) > 0)) {
                     $srcW = $scanX - $srcX;
                  }
               }
            }
         }

         $fontdata{fonts}{$font}{$fontsize}{$char}{width}  = ($fontsize / $fontdata{fonts}{$font}{height}) * $srcW;
         $fontdata{fonts}{$font}{$fontsize}{$char}{height} = $fontsize * 2;

         $fontdata{fonts}{$font}{$fontsize}{$char}{char} = new GD::Image($fontdata{fonts}{$font}{$fontsize}{$char}{width}, $fontdata{fonts}{$font}{$fontsize}{$char}{height}, 1);
         $fontdata{fonts}{$font}{$fontsize}{$char}{char}->transparent($image->colorAllocate(0,0,0));

         $fontdata{fonts}{$font}{$fontsize}{$char}{char}->copyResampled($graphics{GD}{$font}, 0, 0,
                                                                        $srcX  - 1,
                                                                        $charY - ($fontdata{fonts}{$font}{height} / 2),
                                                                        $fontdata{fonts}{$font}{$fontsize}{$char}{width},
                                                                        $fontdata{fonts}{$font}{$fontsize}{$char}{height},
                                                                        $srcW + 2,
                                                                        $fontdata{fonts}{$font}{height} * 2);

      }

      $image->copy($fontdata{fonts}{$font}{$fontsize}{$char}{char},
                   $charpos, 0, 0, 0,
                   $fontdata{fonts}{$font}{$fontsize}{$char}{width},
                   $fontdata{fonts}{$font}{$fontsize}{$char}{height});

      $charpos = $charpos + $fontdata{fonts}{$font}{$fontsize}{$char}{width};

   }

   return $image;

}